@@ -7,6 +7,7 @@ from uuid import UUID
 # https://requests.readthedocs.io
 import requests
 
+from seatable_api.message import get_sender_by_account
 from .constants import ROW_FILTER_KEYS, ColumnTypes
 from .constants import RENAME_COLUMN, RESIZE_COLUMN, FREEZE_COLUMN, MOVE_COLUMN, MODIFY_COLUMN_TYPE, DELETE_COLUMN
 from .socket_io import SocketIO
@@ -71,6 +72,14 @@ class SeaTableAPI(object):
         clone.timeout = self.timeout
         return clone
 
+    def _get_msg_sender_by_account(self, account_name):
+        try:
+            account = self._get_account_detail(account_name)
+            msg_sender = get_sender_by_account(account)
+        except:
+            msg_sender = None
+        return msg_sender
+
     def auth(self, with_socket_io=False):
         """Auth to SeaTable
         """
@@ -95,12 +104,18 @@ class SeaTableAPI(object):
     def _metadata_server_url(self):
         return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/metadata/'
 
+    def _table_server_url(self):
+        return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/tables/'
+
     def _row_server_url(self):
         return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/rows/'
 
     def _batch_row_server_url(self):
         return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/batch-append-rows/'
 
+    def _batch_update_row_server_url(self):
+        return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/batch-update-rows/'
+
     def _batch_delete_row_server_url(self):
         return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/batch-delete-rows/'
 
@@ -110,6 +125,9 @@ class SeaTableAPI(object):
     def _row_link_server_url(self):
         return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/links/'
 
+    def _batch_update_row_link_server_url(self):
+        return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/batch-update-links/'
+
     def _app_download_link_url(self):
         return self.server_url + '/api/v2.1/dtable/app-download-link/'
 
@@ -119,6 +137,35 @@ class SeaTableAPI(object):
     def _column_server_url(self):
         return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/columns/'
 
+    def _column_options_server_url(self):
+        return self.dtable_server_url + '/api/v1/dtables/' + self.dtable_uuid + '/column-options/'
+
+    def _third_party_accounts_url(self):
+        return self.server_url + '/api/v2.1/dtable/third-party-account/'
+
+    def _get_account_detail(self, account_name):
+        url = self._third_party_accounts_url()
+        params = {
+            'account_name': account_name
+        }
+        headers = parse_headers(self.token)
+        response = requests.get(url, params=params, headers=headers, timeout=self.timeout)
+
+        data = parse_response(response)
+        return data.get('account')
+
+    def send_email(self, account_name, msg,  **kwargs):
+        msg_sender = self._get_msg_sender_by_account(account_name)
+        if not msg_sender or msg_sender.msg_type != 'email':
+            raise ValueError('Email message sender does not configered.')
+        msg_sender.send_msg(msg, **kwargs)
+
+    def send_wechat_msg(self, account_name, msg):
+        msg_sender = self._get_msg_sender_by_account(account_name)
+        if not msg_sender or msg_sender.msg_type != 'wechat':
+            raise ValueError('Wechat message sender does not configered.')
+        msg_sender.send_msg(msg)
+
     def get_metadata(self):
         """
         :return: dict
@@ -128,10 +175,27 @@ class SeaTableAPI(object):
         data = parse_response(response)
         return data.get('metadata')
 
-    def list_rows(self, table_name, view_name=None):
+    def add_table(self, table_name, lang='en'):
+        """
+        :param table_name: str
+        :param lang: str, currently 'en' for English, and 'zh-cn' for Chinese
+        """
+        url = self._table_server_url()
+        json_data = {
+            'table_name': table_name,
+            'lang': lang,
+        }
+        response = requests.post(url, json=json_data, headers=self.headers, timeout=self.timeout)
+        return parse_response(response)
+
+    def list_rows(self, table_name, view_name=None, order_by=None, desc=False, start=None, limit=None):
         """
         :param table_name: str
         :param view_name: str
+        :param order_by: str
+        :param desc: boolean
+        :param start: int
+        :param limit: int
         :return: list
         """
         url = self._row_server_url()
@@ -140,6 +204,13 @@ class SeaTableAPI(object):
         }
         if view_name:
             params['view_name'] = view_name
+        if order_by:
+            params['order_by'] = order_by
+            params['direction'] = 'desc' if desc else 'asc'
+        if start:
+            params['start'] = start
+        if limit:
+            params['limit'] = limit
         response = requests.get(url, params=params, headers=self.headers, timeout=self.timeout)
         data = parse_response(response)
         return data.get('rows')
@@ -214,6 +285,20 @@ class SeaTableAPI(object):
         response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
         return parse_response(response)
 
+    def batch_update_rows(self, table_name, rows_data):
+        """
+        :param table_name: str
+        :param rows_data: list
+        :return:
+        """
+        url = self._batch_update_row_server_url()
+        json_data = {
+            'table_name': table_name,
+            'updates': rows_data,
+        }
+        response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
+        return parse_response(response)
+
     def delete_row(self, table_name, row_id):
         """
         :param table_name: str
@@ -342,6 +427,37 @@ class SeaTableAPI(object):
         response = requests.delete(url, json=json_data, headers=self.headers, timeout=self.timeout)
         return parse_response(response)
 
+    def update_link(self, link_id, table_id, other_table_id, row_id, other_rows_ids):
+        """
+        :param link_id: str
+        :param table_id: str
+        :param other_table_id: str
+        :param row_id: str
+        :param other_rows_ids: list
+        """
+        url = self._row_link_server_url()
+        json_data = {
+            'link_id': link_id,
+            'table_id': table_id,
+            'other_table_id': other_table_id,
+            'row_id': row_id,
+            'other_rows_ids': other_rows_ids,
+        }
+        response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
+        return parse_response(response)
+
+    def batch_update_links(self, links):
+        """
+        :param links: list
+        """
+        url = self._batch_update_row_link_server_url()
+        json_data = {
+            'links': links
+        }
+
+        response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
+        return parse_response(response)
+
     def list_columns(self, table_name, view_name=None):
         """
         :param table_name: str
@@ -365,12 +481,13 @@ class SeaTableAPI(object):
                 return column.get('data', {}).get('link_id')
         raise ValueError('link type column "%s" does not exist in current view' % column_name)
 
-    def insert_column(self, table_name, column_name, column_type, column_key=None):
+    def insert_column(self, table_name, column_name, column_type, column_key=None, column_data=None):
         """
         :param table_name: str
         :param column_name: str
         :param column_type: ColumnType enum
         :param column_key: str, which you want to insert after
+        :param column_data: dict, config information of column
         :return: dict
         """
         if column_type not in ColumnTypes:
@@ -382,7 +499,9 @@ class SeaTableAPI(object):
             'column_type': column_type.value
         }
         if column_key:
-            json_data['column_key'] = column_key
+            json_data['anchor_column'] = column_key
+        if column_data:
+            json_data['column_data'] = column_data
         response = requests.post(url, json=json_data, headers=self.headers, timeout=self.timeout)
         data = parse_response(response)
         return data
@@ -398,7 +517,7 @@ class SeaTableAPI(object):
         json_data = {
             'op_type': RENAME_COLUMN,
             'table_name': table_name,
-            'column_key': column_key,
+            'column': column_key,
             'new_column_name': new_column_name
         }
         response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
@@ -417,7 +536,7 @@ class SeaTableAPI(object):
         json_data = {
             'op_type': RESIZE_COLUMN,
             'table_name': table_name,
-            'column_key': column_key,
+            'column': column_key,
             'new_column_width': new_column_width
         }
         response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
@@ -435,7 +554,7 @@ class SeaTableAPI(object):
         json_data = {
             'op_type': FREEZE_COLUMN,
             'table_name': table_name,
-            'column_key': column_key,
+            'column': column_key,
             'frozen': frozen
         }
         response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
@@ -453,8 +572,8 @@ class SeaTableAPI(object):
         json_data = {
             'op_type': MOVE_COLUMN,
             'table_name': table_name,
-            'column_key': column_key,
-            'target_column_key': target_column_key
+            'column': column_key,
+            'target_column': target_column_key
         }
         response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
         data = parse_response(response)
@@ -473,13 +592,29 @@ class SeaTableAPI(object):
         json_data = {
             'op_type': MODIFY_COLUMN_TYPE,
             'table_name': table_name,
-            'column_key': column_key,
+            'column': column_key,
             'new_column_type': new_column_type.value
         }
         response = requests.put(url, json=json_data, headers=self.headers, timeout=self.timeout)
         data = parse_response(response)
         return data
 
+    def add_column_options(self, table_name, column, options):
+        """
+        :param table_name: str
+        :param column: str
+        :param options: list
+        """
+        url = self._column_options_server_url()
+        json_data = {
+            "table_name": table_name,
+            "column": column,
+            "options": options
+        }
+        response = requests.post(url, json=json_data, headers=self.headers, timeout=self.timeout)
+        data = parse_response(response)
+        return data
+
     def delete_column(self, table_name, column_key):
         """
         :param table_name: str
@@ -489,7 +624,7 @@ class SeaTableAPI(object):
         url = self._column_server_url()
         json_data = {
             'table_name': table_name,
-            'column_key': column_key
+            'column': column_key
         }
         response = requests.delete(url, json=json_data, headers=self.headers, timeout=self.timeout)
         data = parse_response(response)
